import os
import logging

from functools import partial
from os.path import dirname, getsize, getmtime, join as pjoin
from subprocess import Popen, STDOUT, PIPE

from tornado import web, websocket, gen, ioloop
from tornado.escape import json_encode, json_decode
from tornado.process import Subprocess

STREAM = Subprocess.STREAM
log = logging.getLogger('logtail')
io_loop = ioloop.IOLoop.instance()


class Commands:
    def __init__(self, grep='grep', awk='gawk', tail='tail'):
        self.grepexe = grep
        self.awkexe = awk
        self.tailexe = tail

    def awk(self, script, fn, stdout, stderr, **kw):
        cmd = [self.awkexe, '--sandbox', script]
        if fn: cmd.append(fn)
        p = Subprocess(cmd, stdout=stdout, stderr=stderr, **kw)
        log.debug('running awk %s, pid: %s', cmd, p.proc.pid)
        return p

    def grep(self, regex, fn, stdout, stderr, **kw):
        cmd = [self.grepexe, '--line-buffered', '--color=never', '-e', regex]
        if fn: cmd.append(fn)
        p = Subprocess(cmd, stdout=stdout, stderr=stderr, **kw)
        log.debug('running grep %s, pid: %s', cmd, p.proc.pid)
        return p

    def tail(self, n, fn, stdout, stderr, **kw):
        cmd = (self.tailexe, '-n', str(n), '-f', fn)
        p = Subprocess(cmd, stdout=stdout, stderr=stderr, **kw)
        log.debug('running tail %s, pid: %s', cmd, p.proc.pid)
        return p

    def tail_awk(self, n, fn, script, stdout, stderr):
        tail = self.tail(n, fn, stdout=PIPE, stderr=STREAM)
        awk = self.awk(script, None, stdout=STREAM, stderr=STREAM, stdin=tail.stdout)
        return tail, awk

    def tail_grep(self, n, fn, regex, stdout, stderr):
        tail = self.tail(n, fn, stdout=PIPE, stderr=STREAM)
        grep = self.grep(regex, None, stdout=STREAM, stderr=STREAM, stdin=tail.stdout)
        tail.stdout.close()
        return tail, grep


class BaseHandler(web.RequestHandler):
    def __init__(self, *args, **kw):
        super(BaseHandler, self).__init__(*args, **kw)
        self.config = self.application.config
        self.cconfig = self.application.cconfig

class Index(BaseHandler):
    def get(self):
        files = self.config['files']['__ungrouped__']
        root = self.config['relative-root']
        files = Files.statfiles(files)
        cconfig = json_encode(self.cconfig)
        self.render('index.html', files=files, root=root, client_config=cconfig)

class Files(BaseHandler):
    @staticmethod
    def statfiles(files):
        for fn in files:
            if not os.access(fn, os.R_OK):
                continue
            yield fn, getsize(fn), getmtime(fn)

    def get(self):
        files = self.config['files']['__ungrouped__']
        files = Files.statfiles(files)
        res = {'files': list(files)}
        self.set_header('Content-Type', 'application/json')
        self.write(json_encode(res))

class Fetch(BaseHandler):
    def error(self, code, msg):
        self.set_header('Content-Type', 'text/html')
        self.set_status(500)
        self.finish(
            '<html><title>%(code)d: %(message)s</title>'
            '<body><tt>%(code)d: %(message)s</tt></body></html>' % \
                {'code': code, 'message': msg})

    def get(self, path):
        if not self.config['allow-transfers']:
            self.error(500, 'transfers not allowed'); return

        if path not in self.config['files']['__ungrouped__']:
            self.error(404, 'file not found'); return

        # basename = os.path.basename(path)
        # self.set_header('Content-Disposition', 'attachment; filename="%s"' % path+'asdf');
        self.set_header('Content-Type', 'text/plain')
        with open(path) as fh:
            self.write(fh.read())  # todo: stream

class WebsocketCommands(websocket.WebSocketHandler):
    def __init__(self, *args, **kw):
        super(WebsocketCommands, self).__init__(*args, **kw)

        self.config = self.application.config
        self.cmd = Commands()
        self.connected = True
        self.tail = None
        self.grep = None
        self.awk = None

    def stdout_callback(self, fn, stream, data):
        log.debug('stdout: %s\n', data.decode('utf8'))
        if not self.connected:
            return

        msg = {fn: data.decode('utf8').splitlines(True)}
        self.wjson(msg)

    def stderr_callback(self, fn, stream, data):
        log.debug('stderr: %s', data)
        if not self.connected:
            return

        text = data.decode('utf8')

        if text.endswith(': file truncated\n'):
            text = 'truncated'
        else:
            text = text.splitlines()

        msg = {'fn': fn, 'err': text}
        self.wjson(msg)

    def killall(self):
        if self.tail:
            log.debug('killing tail process: %s', self.tail.pid)
            self.tail.stdout.close()
            self.tail.stderr.close()
            self.tail.proc.kill()
            self.tail = None

        if self.awk:
            log.debug('killing awk process: %s', self.awk.pid)
            self.awk.stdout.close()
            self.awk.stderr.close()
            self.awk.proc.kill()
            self.awk = None

        if self.grep:
            log.debug('killing grep process: %s', self.grep.pid)
            self.grep.stdout.close()
            self.grep.stderr.close()
            self.grep.proc.kill()
            self.grep = None

    def on_message(self, message):
        msg = json_decode(message)
        log.debug('received message: %s', msg)

        self.killall()

        if 'tail' in msg:
            fn = msg['tail']
            if fn in self.config['files']['__ungrouped__']:
                n = msg.get('last', 10)
                self.tail = self.cmd.tail(n, fn, STREAM, STREAM)

                outcb = partial(self.stdout_callback, fn, self.tail.stdout)
                errcb = partial(self.stderr_callback, fn, self.tail.stderr)
                self.tail.stdout.read_until_close(outcb, outcb)
                self.tail.stderr.read_until_close(errcb, errcb)

        elif 'grep' in msg:
            fn = msg['grep']
            if fn in self.config['files']['__ungrouped__']:
                n = msg.get('last', 10)
                regex = msg.get('script', '.*')

                # self.tail, self.grep = self.cmd.tail_grep2(n, fn, regex)
                self.tail, self.grep = self.cmd.tail_grep(n, fn, regex, STREAM, STREAM)

                outcb = partial(self.stdout_callback, fn, self.grep.stdout)
                errcb = partial(self.stderr_callback, fn, self.grep.stderr)
                # self.tail.stderr.read_until_close(errcb, errcb)
                self.grep.stdout.read_until_close(outcb, outcb)
                self.grep.stderr.read_until_close(errcb, errcb)

        elif 'awk' in msg:
            fn = msg['awk']
            if fn in self.config['files']['__ungrouped__']:
                n = msg.get('last', 10)
                script = msg.get('script', '{print $0}')

                self.tail, self.awk = self.cmd.tail_awk(n, fn, script, STREAM, STREAM)

                outcb = partial(self.stdout_callback, fn, self.awk.stdout)
                errcb = partial(self.stderr_callback, fn, self.awk.stderr)
                # self.tail.stderr.read_until_close(errcb, errcb)
                self.awk.stdout.read_until_close(outcb, outcb)
                self.awk.stderr.read_until_close(errcb, errcb)

    def on_close(self):
        self.killall()
        self.connected = False
        log.debug('connection closed')

    def wjson(self, data):
        return self.write_message(json_encode(data))

class Application(web.Application):
    here = dirname(__file__)

    def __init__(self, config, cconfig={}, template_dir=None, assets_dir=None):
        routes = [
          (r'/assets/(.*)', web.StaticFileHandler, {'path': pjoin(self.here, '../assets/')}),
          (r'/files', Files),
          (r'/fetch/(.*)', Fetch),
          (r'/', Index),
          (r'/ws', WebsocketCommands),
        ]

        if not template_dir: pjoin(self.here, '../templates')
        if not assets_dir:   pjoin(self.here, '../assets')

        log.debug('template dir: %s', template_dir)
        log.debug('static dir: %s', assets_dir)

        settings = { 
          'static_path':   assets_dir,
          'template_path': template_dir,
          'debug': config['debug'],
        }

        super(Application, self).__init__(routes, **settings)
        self.config = config 
        self.cconfig = cconfig



#!/usr/bin/env python3

import os, sys
import glob
import pprint
import logging
import argparse
import textwrap
import collections
import pkg_resources

from tornado import ioloop, httpserver
from tailon.server import Application
from tailon.version import version_verbose


# setup logging
log = logging.getLogger('logtail')
ch = logging.StreamHandler()
ft = logging.Formatter('[+%(relativeCreated)f][%(levelname)5s] %(message)s')

ch.setFormatter(ft)
ch.setLevel(logging.DEBUG)

log.setLevel(logging.INFO)
log.addHandler(ch)
log.propagate = 0

# tornado access logging
weblog = logging.getLogger('tornado.access')
weblog.addHandler(ch)
weblog.setLevel(logging.WARN)
weblog.propagate = 0

# tornado application logging
applog = logging.getLogger('tornado.application')
applog.addHandler(ch)
applog.setLevel(logging.WARN)
applog.propagate = 0


def parseaddr(arg):
    tmp = arg.split(':')
    port = int(tmp[-1])
    addr = ''.join(tmp[:-1])
    addr = '' if addr == '*' else addr
    return port, addr

def parseconfig(cfg):
    import yaml

    raw_config = yaml.load(cfg)

    port, addr = parseaddr(raw_config.get('bind', 'localhost:8080'))
    config = {
        'port': port,
        'addr': addr,
        'debug': raw_config.get('debug', False),
        'allow-transfers': raw_config.get('allow-transfers', False),
        'relative-root':   raw_config.get('relative-root', '/'),
    }

    if 'files' not in raw_config or not len(raw_config['files']):
        raise Exception('missing or empty "files" config entry')

    files = config['files'] = collections.OrderedDict()
    files['__ungrouped__'] = []

    def helper(el, group='__ungrouped__', indict=False):
        for i in el:
            if isinstance(i, dict):
                if indict:
                    raise Exception('more than two sub-levels under "files"')
                name, j = list(i.items())[0]
                helper(j, name, True)
                continue
            for fn in glob.glob(i):
                if not os.access(fn, os.R_OK):
                    log.info('skipping unreadable file: %s', fn)
                    continue
                d = files.setdefault(group, [])
                d.append(fn)

    helper(raw_config['files'])
    return config

def parseopts(args=None):
    description = '''
    Tailon is a webapp for looking at and searching through log files.
    '''

    epilog = '''
    Example config file:
      bind: 0.0.0.0:8080      # address and port to bind on 
      allow-transfers: true   # allow file downloads
      relative-root: /tailon  # web app root path (default: '')
    
      files:
        - '/var/log/messages'
        - '/var/log/nginx/*.log'
        - '/var/log/xorg.[0-10].log'
        - 'cron':             # sub-section
            - '/var/log/cron*'
    '''

    p = argparse.ArgumentParser(formatter_class=CompactHelpFormatter,
                                description=textwrap.dedent(description),
                                epilog=textwrap.dedent(epilog),
                                add_help=False)

    group = p.add_argument_group('Required arguments')
    arg = group.add_argument
    arg('-c', '--config', type=argparse.FileType('r'), metavar='path', help='yaml config file')
    arg('-f', '--files', metavar='path', nargs='+', help='list of files or file wildcards to expose')

    group = p.add_argument_group('Optional arguments')
    arg = group.add_argument
    arg('-h', '--help', action='help', help='show this helm message and exit')
    arg('-d', '--debug', action='store_true', help='show debug messages')
    arg('-v', '--version', action='version', version=version_verbose())
    arg('-b', '--bind', metavar='addr:port', help='listen on the specified address and port')
    arg('-r', '--relative-root', metavar='path', default='', help='web app root path')
    arg('-a', '--allow-transfers', action='store_true', help='allow file downloads')

    return p, p.parse_args(args)
    
def main_config(opts):
    if opts.config:
        config = parseconfig(opts.config)
    else:
        port, addr = parseaddr(opts.bind if opts.bind else 'localhost:8080')
        config = {
            'port': port,
            'addr': addr,
            'files': {'__ungrouped__':[]},
            'allow-transfers': opts.allow_transfers,
            'relative-root': opts.__dict__.get('relative_root', ''),
            'debug': opts.__dict__.get('debug', False),
        }

        for fn in opts.files:
            for fn in glob.glob(fn):
                if not os.access(fn, os.R_OK):
                    log.info('skipping unreadable file: %s', fn)
                    continue
                if fn.startswith('./'):
                    fn = fn.lstrip('./')
                config['files']['__ungrouped__'].append(fn)

    return config

def main(argv=sys.argv):
    parser, opts = parseopts()

    if not opts.config and not opts.files:
        parser.print_help()
        print('\nError: must specify files on the command line or through a config file')
        sys.exit(1)

    if opts.debug:
        log.setLevel(logging.DEBUG)
        applog.setLevel(logging.DEBUG)
        weblog.setLevel(logging.DEBUG)

    try:
        template_dir = pkg_resources.resource_filename('tailon', '../templates')
        assets_dir = pkg_resources.resource_filename('tailon', '../assets')
    except ImportError:
        template_dir, assets_dir = None, None

    config = main_config(opts)

    application = Application(config, {}, template_dir, assets_dir)
    server = httpserver.HTTPServer(application)
    server.listen(config['port'], config['addr'])

    lines = []
    for group, files in config['files'].items():
        group = group if group != '__ungrouped__' else 'ungrouped'
        lines.append('%s:' % group)
        for fn in files: lines.append(fn)
    log.debug('Config:\n%s', pprint.pformat(config))
    log.debug('Files:\n%s', '\n - '.join(lines))
    
    loop = ioloop.IOLoop.instance()
    msg = 'Listening on %s:%s' % (config['addr'], config['port'])
    loop.add_callback(log.info, msg)
    loop.start()
    
class CompactHelpFormatter(argparse.RawTextHelpFormatter):
    def __init__(self, *args, **kw):
        super(CompactHelpFormatter, self).__init__(*args, max_help_position=35, **kw)

    def _format_usage(self, *args, **kw):
        usage = super(CompactHelpFormatter, self)._format_usage(*args, **kw)
        return usage.capitalize()

    def _format_action_invocation(self, action):
        if not action.option_strings:
            default = self._get_default_metavar_for_positional(action)
            metavar, = self._metavar_formatter(action, default)(1)
            return metavar
        else:
            res = ', '.join(action.option_strings)
            default = self._get_default_metavar_for_optional(action)
            args_string = self._format_args(action, default)
            res = '%s %s' % (res, args_string)
            return res


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# encoding: utf-8

'''
Version information constants and auxiliary functions.
'''


VERSION = (0, 1, 0)


import os
import subprocess as sub

__here__ = os.path.abspath(os.path.dirname(__file__))

def _check_output(*cmd):
    p = sub.Popen(cmd, stdout=sub.PIPE, stderr=sub.PIPE, cwd=__here__)
    return p.communicate()[0].rstrip('\n')

_gitsha = lambda : _check_output('git', 'rev-parse',    'HEAD')
_gitbrc = lambda : _check_output('git', 'symbolic-ref', 'HEAD').replace('refs/heads/', '')

def version():
    return '.'.join([str(i) for i in VERSION])

def version_verbose():
    res = 'tailon, version %s' % version()
    try:
        sha = _gitsha() ; brc = _gitbrc()
        if sha or brc:
            res = '%s  (%s:%s)' % (res, brc, sha[:8])
    except:
        pass

    return res


__all__ = (VERSION, version, version_verbose)

if __name__ == '__main__':
    print(version_verbose())



#!/usr/bin/env python
# encoding: utf-8

from os.path import abspath, dirname, join
from setuptools import setup

classifiers = (
    'Development Status :: 3 - Alpha',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.1',
    'Programming Language :: Python :: 3.2',
    'Programming Language :: Python :: 3.3',
    'License :: OSI Approved :: BSD License',
    'Intended Audience :: Developers',
    'Operating System :: POSIX :: Linux',
)

kw = {
    'name'             : 'tailon',
    'version'          : '0.1.0',
    'description'      : 'webapp for looking at and searching through log files',
    'long_description' : open(join(abspath(dirname(__file__)), 'README.rst')).read(),
    'author'           : 'Georgi Valkov',
    'author_email'     : 'georgi.t.valkov@gmail.com',
    'license'          : 'New BSD License',
    'url'              : 'https://github.com/gvalkov/tailon',
    'keywords'         : 'log monitoring tail',
    'classifiers'      : classifiers,
    'packages'         : ['tailon'],
    'install_requires' : ['tornado>=3.0.1', 'PyYAML>=3.10'],
    'entry_points'     : {'console_scripts':['tailon = tailon.main:main']},
    'zip_safe'         : False,
    'package_data'     : {
        'tailon' : ['../assets/js/vendor/*',
                    '../assets/js/main.js',
                    '../assets/css/*',
                    '../assets/favicon.ico',
                    '../assets/fonts/*',
                    '../templates/*',
                    ]
        },
}

setup(**kw)
